from __future__ import annotations

import tokenize_rt


def add_noqa_comments(src: str, lines: set[int], error_code: str) -> str:
    """Add `noqa` comments to some code.

    Args:
        src: The content of the module to add `nqa` comments to.
        lines: The lines on which to add `noqa` coments.
        code: The error code to silence.

    Returns:
        The content of the module with the additional comments added.
    """
    tokens = tokenize_rt.src_to_tokens(src)

    for idx, token in tokenize_rt.reversed_enumerate(tokens):
        if token.line not in lines:
            continue
        if not token.src.strip():
            continue

        if token.name == 'COMMENT':
            new_comment = add_code_to_comment(token.src, error_code)
            tokens[idx] = tokens[idx]._replace(src=new_comment)
        else:
            tokens.insert(
                idx+1, tokenize_rt.Token('COMMENT', f'# noqa: {error_code}'),
            )
            tokens.insert(idx+1, tokenize_rt.Token('UNIMPORTANT_WS', '  '))

        lines.remove(token.line)

    return tokenize_rt.tokens_to_src(tokens)


def add_code_to_comment(comment: str, code: str) -> str:
    """Add to a comment to make it a `noqa` comment.

    If the comment already includes a `noqa` section, this will add the code
    to the list of silenced errors.
    """
    if 'noqa: ' in comment:
        return comment.replace(
            'noqa: ', f'noqa: {code},', 1,
        )
    else:
        return comment + f'  # noqa: {code}'



from __future__ import annotations

import sys
from pathlib import Path

import pytest
from pytest_subprocess import FakeProcess

from silence_lint_error.silence_lint_error import main


def test_main(tmp_path: Path, capsys: pytest.CaptureFixture[str]):
    python_module = tmp_path / 't.py'
    python_module.write_text("""\
import sys
import glob  # noqa: F401
from json import *  # additional comment
from os import *  # noqa: F403
from pathlib import *  # noqa: F403  # additional comment
""")

    ret = main(('flake8', 'F401', str(python_module)))

    assert ret == 1
    assert python_module.read_text() == """\
import sys  # noqa: F401
import glob  # noqa: F401
from json import *  # additional comment  # noqa: F401
from os import *  # noqa: F401,F403
from pathlib import *  # noqa: F401,F403  # additional comment
"""

    captured = capsys.readouterr()
    assert captured.out == f"""\
{python_module}
"""
    assert captured.err == """\
-> finding errors with flake8
found errors in 1 files
-> adding comments to silence errors
"""


def test_main_no_violations(
        tmp_path: Path, capsys: pytest.CaptureFixture[str],
):
    src = """\
def foo():
    print('hello there')
"""

    python_module = tmp_path / 't.py'
    python_module.write_text(src)

    ret = main(('flake8', 'F401', str(python_module)))

    assert ret == 0
    assert python_module.read_text() == src

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> finding errors with flake8
no errors found
"""


def test_not_installed(capsys: pytest.CaptureFixture[str]):
    with FakeProcess() as process:
        process.register(
            (sys.executable, '-mflake8', process.any()),
            returncode=1, stderr='/path/to/python3: No module named flake8\n',
        )

        ret = main(('flake8', 'F401', 'path/to/file.py'))

    assert ret == 1

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> finding errors with flake8
ERROR: /path/to/python3: No module named flake8
"""



from __future__ import annotations

import pytest

from silence_lint_error.noqa import add_code_to_comment
from silence_lint_error.noqa import add_noqa_comments


def test_add_noqa_comments():
    src = """\
# a single-line statement on line 2
foo = 'bar'

# a function on line 5
def baz(
a: int,
b: int,
) -> str:
...

# a multi-line string on line 12
s = '''
hello there
'''
"""

    assert add_noqa_comments(
        src, {2, 5, 12}, 'ABC123',
    ) == """\
# a single-line statement on line 2
foo = 'bar'  # noqa: ABC123

# a function on line 5
def baz(  # noqa: ABC123
a: int,
b: int,
) -> str:
...

# a multi-line string on line 12
s = '''
hello there
'''  # noqa: ABC123
"""


@pytest.mark.parametrize(
    'original, expected', (
        ('something', 'something  # noqa: ABC1'),
        ('noqa: XYZ0', 'noqa: ABC1,XYZ0'),
        ('something  # noqa: XYZ0', 'something  # noqa: ABC1,XYZ0'),
        ('noqa: XYZ0  # something', 'noqa: ABC1,XYZ0  # something'),
        ('noqa: XYZ0  # noqa: UVW3', 'noqa: ABC1,XYZ0  # noqa: UVW3'),
    ),
)
def test_add_code_to_comment(original, expected):
    assert add_code_to_comment(original, 'ABC1') == expected



from __future__ import annotations

import argparse
import json
import re
import subprocess
import sys
from collections import defaultdict
from collections.abc import Sequence
from typing import NamedTuple
from typing import Protocol
from typing import TYPE_CHECKING

import attrs

from . import noqa

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


# Linters
# =======

class Violation(NamedTuple):
    rule_name: RuleName
    lineno: int


@attrs.frozen
class ErrorRunningTool(Exception):
    proc: subprocess.CompletedProcess[str]


class Linter(Protocol):
    name: str

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        ...

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        ...


class Fixit:
    name = 'fixit'

    def __init__(self) -> None:
        self.error_line_re = re.compile(r'^.*?@\d+:\d+ ')

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                sys.executable, '-mfixit',
                '--rules', rule_name,
                'lint', *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named fixit\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[str, list[Violation]] = defaultdict(list)
        for line in proc.stdout.splitlines():
            found_error = self._parse_output_line(line)
            if found_error:
                filename, violation = found_error
                results[filename].append(violation)
            else:  # pragma: no cover
                pass

        return results

    def _parse_output_line(
            self, line: str,
    ) -> tuple[FileName, Violation] | None:
        if not self.error_line_re.match(line):
            return None

        location, violated_rule_name, *__ = line.split(maxsplit=2)
        filename, position = location.split('@', maxsplit=1)
        lineno, *__ = position.split(':', maxsplit=1)

        rule_name_ = violated_rule_name.removesuffix(':')
        return filename, Violation(rule_name_, int(lineno))

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}

        lines = src.splitlines(keepends=True)

        new_lines = []
        for current_lineno, line in enumerate(lines, start=1):
            if current_lineno in linenos_to_silence:
                leading_ws = line.removesuffix(line.lstrip())
                new_lines.append(f'{leading_ws}# lint-fixme: {rule_name}\n')
            new_lines.append(line)

        return ''.join(new_lines)


class Flake8:
    name = 'flake8'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                sys.executable, '-mflake8',
                '--select', rule_name,
                '--format', '%(path)s %(row)s',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named flake8\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[FileName, list[Violation]] = defaultdict(list)
        for line in proc.stdout.splitlines():
            filename_, lineno_ = line.rsplit(maxsplit=1)
            results[filename_].append(Violation(rule_name, int(lineno_)))

        return results

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}
        return noqa.add_noqa_comments(src, linenos_to_silence, rule_name)


class Ruff:
    name = 'ruff'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                sys.executable, '-mruff',
                '--select', rule_name,
                '--output-format', 'json',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named ruff\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        all_violations = json.loads(proc.stdout)
        results: dict[FileName, list[Violation]] = defaultdict(list)
        for violation in all_violations:
            results[violation['filename']].append(
                Violation(
                    rule_name=violation['code'],
                    lineno=violation['location']['row'],
                ),
            )

        return results

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}
        return noqa.add_noqa_comments(src, linenos_to_silence, rule_name)


LINTERS: dict[str, type[Linter]] = {
    'fixit': Fixit,
    'flake8': Flake8,
    'ruff': Ruff,
}


# CLI
# ===

class Context(NamedTuple):
    rule_name: RuleName
    file_names: list[FileName]
    linter: Linter


def _parse_args(argv: Sequence[str] | None) -> Context:
    parser = argparse.ArgumentParser(
        description='Ignore linting errors by adding ignore/fixme comments.',
    )
    parser.add_argument(
        'linter', choices=LINTERS,
        help='The linter for which to ignore errors',
    )
    parser.add_argument('rule_name')
    parser.add_argument('filenames', nargs='*')
    args = parser.parse_args(argv)

    return Context(
        rule_name=args.rule_name,
        file_names=args.filenames,
        linter=LINTERS[args.linter](),
    )


class NoViolationsFound(Exception):
    pass


@attrs.frozen
class MultipleRulesViolated(Exception):
    rule_names: set[RuleName]


def _find_violations(
        linter: Linter, rule_name: RuleName, file_names: Sequence[FileName],
) -> dict[FileName, list[Violation]]:
    violations = linter.find_violations(rule_name, file_names)

    if not violations:
        raise NoViolationsFound

    violation_names = {
        violation.rule_name
        for file_violations in violations.values()
        for violation in file_violations
    }
    if len(violation_names) != 1:
        raise MultipleRulesViolated(violation_names)

    return violations


def _silence_violations(
        linter: Linter, filename: FileName, violations: Sequence[Violation],
) -> bool:
    with open(filename) as f:
        src = f.read()

    src_with_comments = linter.silence_violations(src, violations)

    with open(filename, 'w') as f:
        f.write(src_with_comments)

    return src_with_comments != src


def main(argv: Sequence[str] | None = None) -> int:
    rule_name, file_names, linter = _parse_args(argv)

    print(f'-> finding errors with {linter.name}', file=sys.stderr)
    try:
        violations = _find_violations(linter, rule_name, file_names)
    except ErrorRunningTool as e:
        print(f'ERROR: {e.proc.stderr.strip()}', file=sys.stderr)
        return e.proc.returncode
    except NoViolationsFound:
        print('no errors found', file=sys.stderr)
        return 0
    except MultipleRulesViolated as e:
        print(
            'ERROR: errors found for multiple rules:', sorted(e.rule_names),
            file=sys.stderr,
        )
        return 1
    else:
        print(f'found errors in {len(violations)} files', file=sys.stderr)

    print('-> adding comments to silence errors', file=sys.stderr)
    ret = 0
    for filename, file_violations in violations.items():
        print(filename)
        ret |= _silence_violations(linter, filename, file_violations)

    return ret


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

from pathlib import Path

import pytest

from silence_lint_error.fix_silenced_error import main


def test_main(tmp_path: Path, capsys: pytest.CaptureFixture[str]):
    src = """\
x = None
# lint-fixme: CollapseIsinstanceChecks
isinstance(x, str) or isinstance(x, int)
isinstance(x, bool) or isinstance(x, float)  # lint-fixme: CollapseIsinstanceChecks

def f(x):
    # lint-ignore: CollapseIsinstanceChecks
    return isinstance(x, str) or isinstance(x, int)
"""

    python_module = tmp_path / 't.py'
    python_module.write_text(src)

    ret = main(('fixit', 'fixit.rules:CollapseIsinstanceChecks', str(python_module)))

    assert ret == 0
    assert python_module.read_text() == """\
x = None
isinstance(x, (str, int))
isinstance(x, (bool, float))

def f(x):
    # lint-ignore: CollapseIsinstanceChecks
    return isinstance(x, str) or isinstance(x, int)
"""

    captured = capsys.readouterr()
    assert captured.out == f"""\
{python_module}
"""
    assert captured.err == """\
-> removing comments that silence errors
-> applying auto-fixes with fixit
🛠️  1 file checked, 1 file with errors, 2 auto-fixes available, 2 fixes applied 🛠️
"""


def test_main_no_violations(tmp_path: Path, capsys: pytest.CaptureFixture[str]):
    src = """\
def foo():
    print('hello there')
"""

    python_module = tmp_path / 't.py'
    python_module.write_text(src)

    ret = main(('fixit', 'fixit.rules:CollapseIsinstanceChecks', str(python_module)))

    assert ret == 0
    assert python_module.read_text() == src

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> removing comments that silence errors
no silenced errors found
"""



from __future__ import annotations

import argparse
import subprocess
import sys
from collections.abc import Iterator
from collections.abc import Sequence
from typing import NamedTuple
from typing import Protocol
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


# Linters
# =======

class Violation(NamedTuple):
    rule_name: RuleName
    lineno: int


class Linter(Protocol):
    name: str

    def remove_silence_comments(self, src: str, rule_name: RuleName) -> str:
        ...

    def apply_fixes(
            self, rule_name: RuleName, filenames: Sequence[str],
    ) -> tuple[int, str]:
        ...


class Fixit:
    name = 'fixit'

    def remove_silence_comments(self, src: str, rule_name: RuleName) -> str:
        return ''.join(
            self._remove_comments(
                src.splitlines(keepends=True), rule_name,
            ),
        )

    def _remove_comments(
            self, lines: Sequence[str], rule_name: RuleName,
    ) -> Iterator[str]:
        __, rule_id = rule_name.rsplit(':', maxsplit=1)
        fixme_comment = f'# lint-fixme: {rule_id}'
        for line in lines:
            if line.strip() == fixme_comment:  # fixme comment only
                continue
            elif line.rstrip().endswith(fixme_comment):  # code then fixme comment
                trailing_ws = line.removeprefix(line.rstrip())
                line_without_comment = (
                    line.rstrip().removesuffix(fixme_comment)  # remove comment
                    .rstrip()  # and remove any intermediate ws
                )
                yield line_without_comment + trailing_ws
            else:
                yield line

    def apply_fixes(
            self, rule_name: RuleName, filenames: Sequence[str],
    ) -> tuple[int, str]:
        proc = subprocess.run(
            (
                sys.executable, '-mfixit',
                '--rules', rule_name,
                'fix', '--automatic', *filenames,
            ),
            capture_output=True, text=True,
        )
        return proc.returncode, proc.stderr.strip()


LINTERS: dict[str, type[Linter]] = {
    'fixit': Fixit,
}


# CLI
# ===

class Context(NamedTuple):
    rule_name: RuleName
    file_names: list[FileName]
    linter: Linter


def _parse_args(argv: Sequence[str] | None) -> Context:
    parser = argparse.ArgumentParser(
        description=(
            'Fix linting errors by removing ignore/fixme comments '
            'and running auto-fixes.'
        ),
    )
    parser.add_argument(
        'linter', choices=LINTERS,
        help='The linter to use to fix the errors',
    )
    parser.add_argument('rule_name')
    parser.add_argument('filenames', nargs='*')
    args = parser.parse_args(argv)

    return Context(
        rule_name=args.rule_name,
        file_names=args.filenames,
        linter=LINTERS[args.linter](),
    )


class NoChangesMade(Exception):
    pass


def _unsilence_violations(
        linter: Linter, rule_name: RuleName, filename: FileName,
) -> None:
    with open(filename) as f:
        src = f.read()

    src_without_comments = linter.remove_silence_comments(src, rule_name)

    if src_without_comments == src:
        raise NoChangesMade

    with open(filename, 'w') as f:
        f.write(src_without_comments)


def main(argv: Sequence[str] | None = None) -> int:
    rule_name, file_names, linter = _parse_args(argv)

    print('-> removing comments that silence errors', file=sys.stderr)
    changed_files = []
    for filename in file_names:
        try:
            _unsilence_violations(linter, rule_name, filename)
        except NoChangesMade:
            continue
        else:
            print(filename)
            changed_files.append(filename)

    if not changed_files:
        print('no silenced errors found', file=sys.stderr)
        return 0

    print(f'-> applying auto-fixes with {linter.name}', file=sys.stderr)
    ret, message = linter.apply_fixes(rule_name, changed_files)
    print(message, file=sys.stderr)

    return ret


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import sys
from pathlib import Path

import pytest
from pytest_subprocess import FakeProcess

from silence_lint_error.silence_lint_error import Fixit
from silence_lint_error.silence_lint_error import main
from silence_lint_error.silence_lint_error import Violation


class TestFixit:
    @pytest.mark.parametrize(
        'lines, expected_violations', (
            pytest.param(
                ['t.py@1:2 MyRuleName: the error message'],
                [('t.py', Violation('MyRuleName', 1))],
                id='single-line',
            ),
            pytest.param(
                ['t.py@1:2 MyRuleName: '],
                [('t.py', Violation('MyRuleName', 1))],
                id='no-message',
            ),
            pytest.param(
                [
                    't.py@1:2 MyRuleName: the error message',
                    'which continue over multiple lines',
                    'just like this one does.',
                ],
                [
                    ('t.py', Violation('MyRuleName', 1)),
                    None,
                    None,
                ],
                id='multi-line',
            ),
            pytest.param(
                [
                    't.py@1:2 MyRuleName: ',
                    'the error message on a new line',
                    'which continue over multiple lines',
                    'just like this one does.',
                ],
                [
                    ('t.py', Violation('MyRuleName', 1)),
                    None,
                    None,
                    None,
                ],
                id='multi-line-leading-ws',
            ),
            pytest.param(
                [
                    't.py@1:2 MyRuleName: the error message',
                    'which continue over multiple lines',
                    'just like this one does.',
                    '',
                ],
                [
                    ('t.py', Violation('MyRuleName', 1)),
                    None,
                    None,
                    None,
                ],
                id='multi-line-trailing-ws',
            ),
        ),
    )
    def test_parse_output_line(
            self,
            lines: list[str],
            expected_violations: list[tuple[str, Violation] | None],
    ):
        violations = [
            Fixit()._parse_output_line(line)
            for line in lines
        ]

        assert violations == expected_violations

    def test_find_violations(self, tmp_path: Path):
        python_module = tmp_path / 't.py'
        python_module.write_text(
            """\
x = None
isinstance(x, str) or isinstance(x, int)
""",
        )

        violations = Fixit().find_violations(
            'fixit.rules:CollapseIsinstanceChecks', [str(python_module)],
        )

        assert violations == {
            str(python_module): [
                Violation('CollapseIsinstanceChecks', 2),
            ],
        }


def test_main(tmp_path: Path, capsys: pytest.CaptureFixture[str]):
    python_module = tmp_path / 't.py'
    python_module.write_text(
        """\
x = None
isinstance(x, str) or isinstance(x, int)

def f(x):
    return isinstance(x, str) or isinstance(x, int)
""",
    )

    ret = main(
        ('fixit', 'fixit.rules:CollapseIsinstanceChecks', str(python_module)),
    )

    assert ret == 1
    assert python_module.read_text() == """\
x = None
# lint-fixme: CollapseIsinstanceChecks
isinstance(x, str) or isinstance(x, int)

def f(x):
    # lint-fixme: CollapseIsinstanceChecks
    return isinstance(x, str) or isinstance(x, int)
"""

    captured = capsys.readouterr()
    assert captured.out == f"""\
{python_module}
"""
    assert captured.err == """\
-> finding errors with fixit
found errors in 1 files
-> adding comments to silence errors
"""


def test_main_no_violations(
        tmp_path: Path, capsys: pytest.CaptureFixture[str],
):
    src = """\
def foo():
    print('hello there')
"""

    python_module = tmp_path / 't.py'
    python_module.write_text(src)

    ret = main(
        ('fixit', 'fixit.rules:CollapseIsinstanceChecks', str(python_module)),
    )

    assert ret == 0
    assert python_module.read_text() == src

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> finding errors with fixit
no errors found
"""


def test_main_multiple_different_violations(
        tmp_path: Path, capsys: pytest.CaptureFixture[str],
):
    src = """\
x = None
isinstance(x, str) or isinstance(x, int)

if True:
    pass
"""

    python_module = tmp_path / 't.py'
    python_module.write_text(src)

    ret = main(('fixit', 'fixit.rules', str(python_module)))

    assert ret == 1

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> finding errors with fixit
ERROR: errors found for multiple rules: ['CollapseIsinstanceChecks', 'NoStaticIfCondition']
"""  # noqa: B950


def test_not_installed(capsys: pytest.CaptureFixture[str]):
    with FakeProcess() as process:
        process.register(
            (sys.executable, '-mfixit', process.any()),
            returncode=1, stderr='/path/to/python3: No module named fixit\n',
        )

        ret = main(('fixit', 'fixit.rules', 'path/to/file.py'))

    assert ret == 1

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == """\
-> finding errors with fixit
ERROR: /path/to/python3: No module named fixit
"""
